/* ============================================================================
 * Copyright (c) 2009-2016 BlueQuartz Software, LLC
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of BlueQuartz Software, the US Air Force, nor the names of its
 * contributors may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The code contained herein was partially funded by the following contracts:
 *    United States Air Force Prime Contract FA8650-07-D-5800
 *    United States Air Force Prime Contract FA8650-10-D-5210
 *    United States Prime Contract Navy N00173-07-C-2068
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#pragma once

#include <memory>

#include <string>
#include <vector>

#include "EbsdLib/Core/EbsdDataArray.hpp"
#include "EbsdLib/EbsdLib.h"

/**
 * @struct PoleFigureConfiguration_t
 * @brief This structure controls how Pole Figures are generated. The Order member
 * is the trickiest to understad. Each SpaceGroup has a default order for displaying
 * the 3 pole figures that are created. If you need a different order then the
 * developer needs to create a 3 element std::vector and set the order index for the
 * Pole Figure. For example the OrthorhombicOps generates 3 Pole Figures in the order
 * of <001>, <100>, <010>. If the developer would like the Pole Figures to appear
 * in the order of <100>, <010>, <100> then the oder needs to be set as [2, 0, 1].
 *
 * The other interesting item is the Labels. Each LaueOps subclass uses a default
 * label for each Pole Figure. If the developer would like to over ride those labels
 * then this member can be set with a 3 Element std::vector<std::string> with the new labels.
 * Note that the new lables will REPLACE the default labels.
 */
struct PoleFigureConfiguration_t
{
  EbsdLib::FloatArrayType* eulers; ///<* The Euler Angles (in Radians) to use for the pole figure
  int imageDim;                    ///<* The height/width of the generated pole figure
  int lambertDim;                  ///<* The dimensions in voxels of the Lambert Square used for interpolation
  int numColors;                   ///<* The number of colors to use in the Pole figure
  double minScale;                 ///<* The minimum scale of the Pole Figure
  double maxScale;                 ///<* The maximum scale of the Pole Figure
  float sphereRadius;              ///<* The radius of the Sphere to compute XYZ coords. Should ALWAYS be 1.0
  bool discrete;                   ///<* Should the Pole Figure be generated as a discrete plot
  bool discreteHeatMap;            ///<* Should the discrete be colored via a heat map style coloring
  std::string colorMap;            ///<* Name of the ColorMap to use
  std::vector<std::string> labels; ///<* The labels for each of the 3 Pole Figures
  std::vector<unsigned int> order; ///<* The order that the pole figures should appear in.
  std::string phaseName;           ///<* The Names of the phase
};

/**
 * @class PoleFigureUtilities PoleFigureUtilities.h /Utilities/PoleFigureUtilities.h
 * @brief This class has functions that help create pole figures.
 * @author Mike Jackson for BlueQuartz Software
 * @date Aug 1 2013
 * @version 1.0
 */
class EbsdLib_EXPORT PoleFigureUtilities
{
public:
  PoleFigureUtilities();
  virtual ~PoleFigureUtilities();

  /**
   * @brief CreateColorImage
   * @param data
   * @param width
   * @param height
   * @param nColors
   * @param name
   * @param min
   * @param max
   * @return
   */
  static EbsdLib::UInt8ArrayType::Pointer CreateColorImage(EbsdLib::DoubleArrayType* data, int width, int height, int nColors, const std::string& name, double min, double max);

  /**
   * @brief CreateColorImage
   * @param data
   * @param config
   * @param image
   */
  static void CreateColorImage(EbsdLib::DoubleArrayType* data, PoleFigureConfiguration_t& config, EbsdLib::UInt8ArrayType* image);

private:
  /**
   * @brief GenerateHexPoleFigures
   * @param eulers
   * @param lambertDimension
   * @param poleFigureDim
   * @param intensity0001 [output]
   * @param intensity1010 [output]
   * @param intensity1120 [output]
   */
  void GenerateHexPoleFigures(EbsdLib::FloatArrayType* eulers, int lambertDimension, int poleFigureDim, EbsdLib::DoubleArrayType::Pointer& intensity0001,
                              EbsdLib::DoubleArrayType::Pointer& intensity1010, EbsdLib::DoubleArrayType::Pointer& intensity1120);

  /**
   * @brief GenerateHexPoleFigures
   * @param eulers
   * @param lambertDimension
   * @param poleFigureDim
   * @param intensity100 [output]
   * @param intensity010 [output]
   * @param intensity001 [output]
   */
  void GenerateOrthoPoleFigures(EbsdLib::FloatArrayType* eulers, int lambertDimension, int poleFigureDim, EbsdLib::DoubleArrayType::Pointer& intensity100,
                                EbsdLib::DoubleArrayType::Pointer& intensity010, EbsdLib::DoubleArrayType::Pointer& intensity001);

public:
  PoleFigureUtilities(const PoleFigureUtilities&) = delete;            // Copy Constructor Not Implemented
  PoleFigureUtilities(PoleFigureUtilities&&) = delete;                 // Move Constructor Not Implemented
  PoleFigureUtilities& operator=(const PoleFigureUtilities&) = delete; // Copy Assignment Not Implemented
  PoleFigureUtilities& operator=(PoleFigureUtilities&&) = delete;      // Move Assignment Not Implemented
};

/**
 * @brief The GeneratePoleFigureRgbaImageImpl class is a wrapper around generating the RGBA image (2D UChar Array with 4 Components) from the
 * intensity image. This should be called from a TBB Task
 */
class GeneratePoleFigureRgbaImageImpl
{
public:
  GeneratePoleFigureRgbaImageImpl(EbsdLib::DoubleArrayType* intensity, PoleFigureConfiguration_t* config, EbsdLib::UInt8ArrayType* rgba);
  virtual ~GeneratePoleFigureRgbaImageImpl();

  void operator()() const;

protected:
  GeneratePoleFigureRgbaImageImpl();

private:
  EbsdLib::DoubleArrayType* m_Intensity = nullptr;
  PoleFigureConfiguration_t* m_Config = nullptr;
  EbsdLib::UInt8ArrayType* m_Rgba = nullptr;
};
